r"""
A tiling Opendap-aware caching proxy server.

This program is a proxy to any Opendap server. Data requests (dods,
ascii and other configurable responses) are cached locally by an
Opendap-aware algorithm; this means that requests for a subset of
data that was already requested, eg, will be read from cache even
though the URLs may be completely different.

The caching mechanism works by splitting an n-dimensional dataset
in pairs of tiles, like a binary k-d tree::

        A 
       / \
      B   C
     /|   |\
    D E   F G

When a slice of the data is requested we check each tile to see if
the data is present in that tile. The result may be this, eg::

        1 
       / \
      1   1
     /|   |\
    1 1   1 0

In order to download the data with the least transferred bytes we
need to request tiles B and F from the Opendap server, store it
locally, and then return the requested slice (which now is contained
in the cached data). If the node E is already in the cache we can
request D instead of B, to minimize downloaded data. 

(c) 2011-2012 Roberto De Almeida, Guilherme Castelao

"""
from __future__ import division

import sys, logging
logging.basicConfig(stream=sys.stdout)
logger = logging.getLogger('pydap')
logger.setLevel(logging.INFO)

import os
from urlparse import urljoin
from collections import defaultdict
from rfc822 import parsedate
import time
from stat import ST_MTIME

from webob import Request
from paste.proxy import Proxy
import requests
import h5py
import numpy as np

from pydap.client import open_url
from pydap.handlers.lib import SimpleHandler
from pydap.lib import walk, combine_slices, fix_slice, hyperslab
from pydap.model import *
from pydap.proxy import ArrayProxy
from pydap.util.rwlock import ReadWriteLock

import pydap.lib
pydap.lib.USER_AGENT = 'DapCache/0.1'


TILESIZE = int(2e6)  # bytes
MAXSIZE  = int(1e8)
LOCK = defaultdict(ReadWriteLock)

# TODO:
# improve storage
# check if cache has exceeded its size: we need to use redis for the RW lock, shared with a script that deletes old files.
#   https://github.com/andymccurdy/redis-py
#   http://chris-lamb.co.uk/2010/06/07/distributing-locking-python-and-redis/
#   http://degizmo.com/2010/03/22/getting-started-redis-and-python/
#   http://redis.io/commands/setnx


class DapCache(object):
    def __init__(self, url, responses, cachedir, tilesize=TILESIZE, maxsize=MAXSIZE):
        self.url = url
        self.responses = responses
        if not os.path.exists(cachedir):
            os.mkdir(cachedir)
        self.cachedir = cachedir
        self.tilesize = int(tilesize)
        self.maxsize = int(maxsize)

    def __call__(self, environ, start_response):
        req = Request(environ)
        if '.' in req.path_info:
            basename, response = req.path_info.rsplit('.', 1)
        else:
            basename, response = req.path_info, None

        # cache a local copy
        if response in self.responses:
            url = urljoin(self.url, basename)
            dataset = open_url(url)
            cachepath = os.path.join(self.cachedir, basename.replace('/', '_'))

            # here we an use mstat to check the mtime of the file, and 
            # do a HEAD on the dataset to compare with Last-Modified header
            r = requests.head(url + '.dods')
            if 'last-modified' in r.headers:
                last_modified = time.mktime(parsedate(r.headers['last-modified']))
                mtime = time.mktime(time.localtime( os.stat(cachepath)[ST_MTIME] ))
                if last_modified > mtime:
                    os.unlink(cachepath)

            # replace data with a caching version
            for var in walk(dataset, BaseType):
                var.data = CachingArrayProxy(
                        cachepath, self.tilesize, self.maxsize,
                        var.type, var.id,
                        var.data.url, var.data.shape, var.data._slice)
            #for var in walk(dataset, SequenceType):
            #    var.data = CachingSequenceProxy(
            #            cachepath,
            #            var.id,
            #            var.data.url, var.data.slice, var.data._slice, var.data.children)

            app = SimpleHandler(dataset)
        # pass this upstream
        else:
            app = Proxy(self.url)

        return app(environ, start_response)


class CachingArrayProxy(ArrayProxy):
    def __init__(self, cachepath, tilesize, maxsize, dtype, id_, url, shape, slice_=None):
        super(CachingArrayProxy, self).__init__(id_, url, shape, slice_)

        self.tilesize = tilesize
        self.maxsize = maxsize

        self.lock = LOCK[cachepath]
        self.fp = h5py.File(cachepath, 'a')

        # how many tiles are we using for the cache? we can calculate
        # this analytically, but this way is more explicit... ;)
        size = np.prod(shape) * dtype.size
        divisions = 0
        while size > tilesize:
            size = np.ceil(size/2)
            divisions += 1
        self.tiles = 2**divisions

        # open cache data/index and create arrays if necessary
        with self.lock.readlock:
            if id_ not in self.fp:
                with self.lock.writelock:
                    self.fp.create_dataset(id_, shape, dtype.typecode)
            if 'index' not in self.fp:
                with self.lock.writelock:
                    self.fp.create_group('index')
            if id_ not in self.fp['index']:
                with self.lock.writelock:
                    self.fp['index'].create_dataset(id_, (self.tiles,), bool)
        self.cache = self.fp[id_]
        self.index = self.fp['index'][id_]

    def __getitem__(self, index):
        """
        Download data for all the tiles containing the request.

        """
        slice_ = combine_slices(self._slice, fix_slice(index, self.shape))
        requested = self.parse_request(slice_)
        with self.lock.readlock:
            needed = requested & ~self.index[:]

            # update cache with needed data
            with self.lock.writelock:
                for tile in self.get_tiles(needed):
                    self.cache[tile] = super(CachingArrayProxy, self).__getitem__(tile)
                # update index with newly requested data
                self.index[:] = self.index[:] | needed

            return self.cache[slice_]

    def parse_request(self, slice_):
        """
        Parse a slice request into an array of requested tiles.

        """
        # check which of the smaller tiles are in the request
        height = int(np.log2(self.tiles))
        requested = []
        queue = [ (0, list(0 for dim in self.shape), list(self.shape)) ]
        while queue:
            depth, start, end = queue.pop(0)
            if depth < height:
                left, right = split(start, end)
                queue.append( ((depth+1,) + left) )
                queue.append( ((depth+1,) + right) )
            else:
                for s, min_, max_ in zip(slice_, start, end):
                    if s.stop <= min_ or s.start >= max_:
                        requested.append(0)
                        break
                else:
                    requested.append(1)

        return np.array(requested, dtype=bool)

    def get_tiles(self, needed):
        r"""
        Combine needed tiles and return them as slices.

        We create a k-d tree and go from the top to the bottom.
        If for a given node all its lower children are needed, say::

                0 
               / \
              0   0
             /|   |\
            1 1   1 1
            
        We can make a request for a continuous region that contains
        all child nodes.

        """
        queue = [ (needed, list(0 for dim in self.shape), list(self.shape)) ]
        while queue:
            needed, start, end = queue.pop(0)
            size = sum(needed) * self.tilesize
            if all(needed) and size < self.maxsize:
                yield tuple(slice(min_, max_, 1) for min_, max_ in zip(start, end))
            elif any(needed):
                middle = np.floor(len(needed) / 2)
                left, right = split(start, end)
                queue.append( ((needed[:middle],) + left) )
                queue.append( ((needed[middle:],) + right) )


def split(start, end):
    # split along longest dimension
    shape = tuple(p1-p0 for p0, p1 in zip(start, end))
    axis = shape.index(max(shape))
    middle = int(start[axis] + np.floor(shape[axis] / 2))

    left = start, end[:axis] + [middle] + end[axis+1:]
    right = start[:axis] + [middle] + start[axis+1:], end

    return left, right


def make_cache(global_conf, url, responses, **kwargs):
    from paste.deploy.converters import aslist
    responses = aslist(responses)
    return DapCache(url, responses, **kwargs)


if __name__ == '__main__':
    app = DapCache('http://pydap.zenofdata.com/', ['dods', 'asc', 'ascii'], '.cache', 64)
    from paste import httpserver
    httpserver.serve(app, '127.0.0.1', port=8003)



"""
BSD Licence
Copyright (c) 2009, Science & Technology Facilities Council (STFC)
All rights reserved.

Patch pydap to enable ESGF security.

@author: Stephen Pascoe <Stephen.Pascoe@stfc.ac.uk>

"""

import urllib2
import re
import os, sys

import M2Crypto

import pydap.lib
from pydap.exceptions import ClientError

import logging
log = logging.getLogger('esgf.security')

def install_esgf_client(certfile, keyfile):
    """
    Based on the CAS authentication example at
    http://pydap.org/client.html#authentication

    This function will patch the pydap package to provide a client
    certificate for HTTPS calls.

    :param certfile: filename of a client certificate or certificate
        chain in PEM format.
    :param keyfile: filename of a private key in PEM format.

    """

    # Create HTTPSHandler    
    ssl_context = M2Crypto.SSL.Context()
    ssl_context.load_cert_chain(certfile, keyfile)
    opener = M2Crypto.m2urllib2.build_opener(ssl_context, 
                                             urllib2.HTTPCookieProcessor,
                                             urllib2.ProxyHandler)

    opener.addheaders = [('User-agent', pydap.lib.USER_AGENT)]
    urllib2.install_opener(opener)

    def new_request(url):
        log.debug('Opening URL %s' % url)

        r = urllib2.urlopen(url.rstrip('?&'))

        resp = r.headers.dict
        resp['status'] = str(r.code)
        data = r.read()

        # When an error is returned, we parse the error message from the
        # server and return it in a ``ClientError`` exception.
        if resp.get("content-description") == "dods_error":
            m = re.search('code = (?P<code>\d+);\s*message = "(?P<msg>.*)"',
                    data, re.DOTALL | re.MULTILINE)
            msg = 'Server error %(code)s: "%(msg)s"' % m.groupdict()
            raise ClientError(msg)

        return resp, data

    from pydap.util import http
    http.request = new_request





"""
A simple file-based Opendap server.

Serves files from a root directory, handling those recognized by installed
handlers.

"""

import re
import os
from os.path import getmtime, getsize
import time
from email.utils import formatdate

from paste.request import construct_url
from paste.httpexceptions import HTTPNotFound, HTTPSeeOther
from paste.fileapp import FileApp

from pydap.handlers.lib import get_handler, load_handlers
from pydap.lib import __version__
from pydap.exceptions import ExtensionNotSupportedError
from pydap.util.template import FileLoader, GenshiRenderer


class FileServer(object):
    def __init__(self, root, templates='templates', catalog='catalog.xml', **config):
        self.root = root.replace('/', os.path.sep)
        self.catalog = catalog
        self.config = config

        loader = FileLoader(templates)
        self.renderer = GenshiRenderer(
                options={}, loader=loader)

        self.handlers = load_handlers()

    def __call__(self, environ, start_response):
        path_info = environ.get('PATH_INFO', '')
        filepath = os.path.abspath(os.path.normpath(os.path.join(
                self.root,
                path_info.lstrip('/').replace('/', os.path.sep))))
        basename, extension = os.path.splitext(filepath)
        assert filepath.startswith(self.root)  # check for ".." exploit

        # try to set our renderer as the default, it none exists
        environ.setdefault('pydap.renderer', self.renderer) 

        # check for regular file or dir request
        if os.path.exists(filepath):
            if os.path.isfile(filepath):
                # return file download
                return FileApp(filepath)(environ, start_response)
            else:
                # return directory listing
                if not path_info.endswith('/'):
                    environ['PATH_INFO'] = path_info + '/'
                    return HTTPSeeOther(construct_url(environ))(environ, start_response)
                return self.index(environ, start_response,
                        'index.html', 'text/html')
        # else check for opendap request
        elif os.path.exists(basename):
            # Update environ with configuration keys (environ wins in case of conflict).
            for k in self.config:
                environ.setdefault(k, self.config[k])
            handler = get_handler(basename, self.handlers)
            return handler(environ, start_response)
        # check for catalog
        elif path_info.endswith('/%s' % self.catalog):
            environ['PATH_INFO'] = path_info[:path_info.rfind('/')]
            return self.index(environ, start_response,
                    'catalog.xml', 'text/xml')
        else:
            return HTTPNotFound()(environ, start_response)

    def index(self, environ, start_response, template_name, content_type):
        # Return directory listing.
        path_info = environ.get('PATH_INFO', '')
        directory = os.path.abspath(os.path.normpath(os.path.join(
                self.root,
                path_info.lstrip('/').replace('/', os.path.sep))))

        mtime = getmtime(directory)
        dirs_ = []
        files_ = []
        for root, dirs, files in os.walk(directory):
            filepaths = [ 
                    os.path.abspath(os.path.join(root, filename))
                    for filename in files ]
            # Get Last-modified.
            filepaths = filter(os.path.exists, filepaths)  # remove broken symlinks
            if filepaths:
                mtime = max(mtime, *map(getmtime, filepaths))

            # Add list of files and directories.
            if root == directory:
                dirs_ = [d for d in dirs if not d.startswith('.')]
                files_ = [{
                        'name': os.path.split(filepath)[1],
                        'size': format_size(getsize(filepath)),
                        'modified': time.localtime(getmtime(filepath)),
                        'supported': supported(filepath, self.handlers),
                        } for filepath in filepaths]

        # Sort naturally using Ned Batchelder's algorithm.
        dirs_.sort(key=alphanum_key)
        files_.sort(key=lambda l: alphanum_key(l['name']))

        # Base URL.
        location = construct_url(environ, with_query_string=False)
        root = construct_url(environ, with_query_string=False, with_path_info=False).rstrip('/')

        context = {
                'environ': environ,
                'root': root,
                'location': location,
                'title': 'Index of %s' % (environ.get('PATH_INFO') or '/'),
                'dirs' : dirs_,
                'files': files_,
                'catalog': self.catalog,
                'version': '.'.join(str(d) for d in __version__)
        }
        template = environ['pydap.renderer'].loader(template_name)
        output = environ['pydap.renderer'].render(template, context, output_format=content_type)
        last_modified = formatdate(time.mktime(time.localtime(mtime)))
        headers = [('Content-type', content_type), ('Last-modified', last_modified)]
        start_response("200 OK", headers)
        return [output.encode('utf-8')]


def supported(filepath, handlers):
    try:
        get_handler(filepath, handlers)
        return True
    except ExtensionNotSupportedError:
        return False


# http://svn.colorstudy.com/home/ianb/ImageIndex/indexer.py
def format_size(size):
    if not size:
        return 'empty'
    if size > 1024:
        size = size / 1024.
        if size > 1024:
            size = size / 1024.
            return '%.1i MB' % size
        return '%.1f KB' % size
    return '%i bytes' % size


def alphanum_key(s):
    """
    Turn a string into a list of string and number chunks.

        >>> alphanum_key("z23a")
        ['z', 23, 'a']

    From http://nedbatchelder.com/blog/200712.html#e20071211T054956

    """
    def tryint(s):
        try:
            return int(s)
        except:
            return s
    return [tryint(c) for c in re.split('([0-9]+)', s)]


def make_app(global_conf, root, templates, **kwargs):
    return FileServer(root, templates=templates, **kwargs)
